@@ -5,17 +5,18 @@
 //! available, using the Time Stamp Counter feature found on modern CPUs.
 //!
 //! # Design
+//!
 //! Internally, `quanta` maintains the concept of two potential clock sources: a reference clock and
 //! a source clock.
 //!
 //! The reference clock is provided by the OS, and always available.  It is equivalent to what is
-//! provided by the standard library in terms of the underlying system calls being made.  As it
-//! uses the native timing facilities provided by the operating system, we ultimately depend on the
-//! OS itself to give us a stable and correct value.
+//! provided by the standard library in terms of the underlying system calls being made.  As it uses
+//! the native timing facilities provided by the operating system, we ultimately depend on the OS
+//! itself to give us a stable and correct value.
 //!
-//! The source clock is a potential clock source based on the Time Stamp Counter feature found on
-//! modern CPUs.  If the TSC feature is not present or is not reliable enough, `quanta` will
-//! transparently utilize the reference clock instead.
+//! The source clock is a potential clock source based on the [Time Stamp Counter][tsc] feature
+//! found on modern CPUs.  If the TSC feature is not present or is not reliable enough, `quanta`
+//! will transparently utilize the reference clock instead.
 //!
 //! Depending on the underlying processor(s) in the system, `quanta` will figure out the most
 //! accurate/efficient way to calibrate the source clock to the reference clock in order to provide
@@ -24,45 +25,62 @@
 //! Details on TSC support, and calibration, are detailed below.
 //!
 //! # Features
-//! Beyond simply taking measurements of the current time, `quanta` provides features for more easily
-//! working with clocks, as well as being able to enhance performance further:
+//!
+//! Beyond simply taking measurements of the current time, `quanta` provides features for more
+//! easily working with clocks, as well as being able to enhance performance further:
 //! - `Clock` can be mocked for testing
 //! - globally accessible "recent" time with amortized overhead
 //!
+//! ## Mocked time
+//!
 //! For any code that uses a `Clock`, a mocked version can be substituted.  This allows for
 //! application authors to control the time in tests, which allows simulating not only the normal
 //! passage of time but provides the ability to warp time forwards and backwards in order to test
 //! corner cases in logic, etc.  Creating a mocked clock can be acheived with [`Clock::mock`], and
 //! [`Mock`] contains more details on mock usage.
 //!
+//! ## Coarsely-updated, or recent, time
+//!
 //! `quanta` also provides a "recent" time feature, which allows a slightly-delayed version of time
 //! to be provided to callers, trading accuracy for speed of access.  An upkeep thread is spawned,
 //! which is responsible for taking measurements and updating the global recent time. Callers then
-//! can access the cached value by calling `Clock::recent`.  This interface can be 4-10x faster
-//! than directly calling `Clock::now`, even when TSC support is available.  As the upkeep thread
-//! is the only code updating the recent time, the accuracy of the value given to callers is
-//! limited by how often the upkeep thread updates the time, thus the trade off between accuracy
-//! and speed of access.
+//! can access the cached value by calling `Clock::recent`.  This interface can be 4-10x faster than
+//! directly calling `Clock::now`, even when TSC support is available.  As the upkeep thread is the
+//! only code updating the recent time, the accuracy of the value given to callers is limited by how
+//! often the upkeep thread updates the time, thus the trade off between accuracy and speed of
+//! access.
 //!
 //! # Feature Flags
-//! `quanta` comes with feature flags that enable convenient conversions to time types in
-//! other popular crates, such as:
+//!
+//! `quanta` comes with feature flags that enable convenient conversions to time types in other
+//! popular crates, such as:
 //! - `prost` - provides an implementation into [`Timestamp`][prost_types_timestamp] from
-//! `prost_types`
+//!   `prost_types`
 //!
 //! # Platform Support
+//!
 //! At a high level, `quanta` carries support for most major operating systems out of the box:
-//! - Windows ([QueryPerformanceCounter])
-//! - macOS/OS X/iOS ([mach_continuous_time])
-//! - Linux/*BSD/Solaris ([clock_gettime])
+//! - Windows ([`QueryPerformanceCounter`][QueryPerformanceCounter])
+//! - macOS/OS X/iOS ([`mach_absolute_time`][mach_absolute_time])
+//! - Linux/*BSD/Solaris ([`clock_gettime`][clock_gettime])
 //!
 //! These platforms are supported in the "reference" clock sense, and support for using the Time
 //! Stamp Counter as a clocksource is more subtle, and explained below.
 //!
-//! # Time Stamp Counter support
-//! Accessing the TSC requires being on the x86_64 architecture, with access to SSE2. Additionally,
-//! the processor must support either constant or nonstop/invariant TSC.  This ensures that the TSC
-//! ticks at a constant rate which can be easily scaled.
+//! ## WASM support
+//!
+//! This library can be built for WASM targets, but in this case the resolution and accuracy of
+//! measurements can be limited by the WASM environment. In particular, when running on the
+//! `wasm32-unknown-unknown` target in browsers, `quanta` will use [windows.performance.now] as a
+//! clock. This mean the accuracy is limited to milliseconds instead of the usual nanoseconds on
+//! other targets. When running within a WASI environment (target `wasm32-wasi`), the accuracy of
+//! the clock depends on the VM implementation.
+//!
+//! # TSC Support
+//!
+//! Accessing the TSC requires being on the `x86_64` architecture, with access to SSE2.
+//! Additionally, the processor must support either constant or nonstop/invariant TSC.  This ensures
+//! that the TSC ticks at a constant rate which can be easily scaled.
 //!
 //! A caveat is that "constant" TSC doesn't account for all possible power states (levels of power
 //! down or sleep that a CPU can enter to save power under light load, etc) and so a constant TSC
@@ -73,68 +91,64 @@
 //! recalibration. Nonstop/invariant TSC does not have this limitation and is stable over long
 //! periods of time.
 //!
-//! Roughly speaking, the following list contains the beginning model/generation of processors
-//! where you should be able to expect having invariant TSC support:
+//! Roughly speaking, the following list contains the beginning model/generation of processors where
+//! you should be able to expect having invariant TSC support:
 //! - Intel Nehalem and newer for server-grade
 //! - Intel Skylake and newer for desktop-grade
 //! - VIA Centaur Nano and newer (circumstantial evidence here)
 //! - AMD Phenom and newer
 //!
-//! Ultimately, `quanta` will query CPUID information to determine if the processor has the
-//! required features to use the TSC.
+//! Ultimately, `quanta` will query CPUID information to determine if the processor has the required
+//! features to use the TSC.
 //!
 //! # Calibration
+//!
 //! As the TSC doesn't necessarily tick at reference scale -- i.e. one tick isn't always one
 //! nanosecond -- we have to apply a scaling factor when converting from source to reference time
-//! scale.  We acquire this scaling factor by querying the processor or calibrating our source
-//! clock to the reference clock.
-//!
-//! In some cases, on newer processors, the frequency of the TSC can be queried directly, providing
-//! a fixed scaling factor with no further calibration necessary.  In other cases, `quanta` will
-//! have to run its own calibration before the clock is ready to be used: repeatedly taking
-//! measurements from both the reference and source clocks until a stable scaling factor has been
-//! established.
+//! scale to provide this.  We acquire this scaling factor by repeatedly taking measurements from
+//! both the reference and source clocks, until we have a statistically-relevant measure of the
+//! average scaling factor.  We do some additional work to convert this scaling factor into a
+//! power-of-two number that allows us to optimize the code, and thus reduce the generated
+//! instructions required to scale a TSC value.
 //!
 //! This calibration is stored globally and reused.  However, the first `Clock` that is created in
-//! an application will block for a small period of time as it runs this calibration loop.  The
-//! time spent in the calibration loop is limited to 200ms overall.  In practice, `quanta` will
-//! reach a stable calibration quickly (usually 10-20ms, if not less) and so this deadline is
-//! unlikely to be reached.
+//! an application will block for a small period of time as it runs this calibration loop.  The time
+//! spent in the calibration loop is limited to 200ms overall.  In practice, `quanta` will reach a
+//! stable calibration quickly (usually 10-20ms, if not less) and so this deadline is unlikely to be
+//! reached.
 //!
 //! # Caveats
+//!
 //! Utilizing the TSC can be a tricky affair, and so here is a list of caveats that may or may not
 //! apply, and is in no way exhaustive:
 //! - CPU hotplug behavior is undefined
 //! - raw values may time warp
 //! - measurements from the TSC may drift past or behind the comparable reference clock
 //!
-//! # WASM support
-//!
-//! This library can be built for WASM targets, but in this case the resolution
-//! and accuracy of measurements can be limited by the WASM environment. In
-//! particular, when running on the `wasm32-unknown-unknown` target in browsers,
-//! `quanta` will use [windows.performance.now] as a clock. This mean the
-//! accuracy is limited to milliseconds instead of the usual nanoseconds on
-//! other targets. When running within a WASI environment (target
-//! `wasm32-wasi`), the accuracy of the clock depends on the VM implementation.
+//! Another important caveat is that `quanta` does not track time across system suspends.  Simply
+//! put, if a time measurement (such as using [`Instant::now`][crate::Instant::now]) is taken, and
+//! then the system is suspended, and then another measurement is taken, the difference between
+//! those the two would not include the time the system was in suspend.
 //!
+//! [tsc]: https://en.wikipedia.org/wiki/Time_Stamp_Counter
 //! [QueryPerformanceCounter]: https://msdn.microsoft.com/en-us/library/ms644904(v=VS.85).aspx
-//! [mach_continuous_time]: https://developer.apple.com/documentation/kernel/1646199-mach_continuous_time
+//! [mach_absolute_time]: https://developer.apple.com/documentation/kernel/1462446-mach_absolute_time
 //! [clock_gettime]: https://linux.die.net/man/3/clock_gettime
 //! [prost_types_timestamp]: https://docs.rs/prost-types/0.7.0/prost_types/struct.Timestamp.html
 //! [windows.performance.now]: https://developer.mozilla.org/en-US/docs/Web/API/Performance/now
+#![deny(missing_docs)]
+#![deny(clippy::all)]
+#![allow(clippy::must_use_candidate)]
+
 use crossbeam_utils::atomic::AtomicCell;
 use std::time::Duration;
 use std::{cell::RefCell, sync::Arc};
 
 use once_cell::sync::OnceCell;
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-use raw_cpuid::CpuId;
 
-mod monotonic;
-use self::monotonic::Monotonic;
-mod counter;
-use self::counter::Counter;
+mod clocks;
+use self::clocks::{Counter, Monotonic};
+mod detection;
 mod mock;
 pub use self::mock::{IntoNanoseconds, Mock};
 mod instant;
@@ -144,12 +158,16 @@ pub use self::upkeep::{Error, Handle, Upkeep};
 mod stats;
 use self::stats::Variance;
 
+// Global clock, used by `Instant::now`.
 static GLOBAL_CLOCK: OnceCell<Clock> = OnceCell::new();
 
+// Global recent measurement, used by `Clock::recent` and `Instant::recent`.
 static GLOBAL_RECENT: AtomicCell<u64> = AtomicCell::new(0);
 
+// Global calibration, shared by all clocks.
 static GLOBAL_CALIBRATION: OnceCell<Calibration> = OnceCell::new();
 
+// Per-thread clock override, used by `quanta::with_clock`, `Instant::now`, and sometimes `Instant::recent`.
 thread_local! {
     static CLOCK_OVERRIDE: RefCell<Option<Clock>> = RefCell::new(None);
 }
@@ -161,13 +179,12 @@ const MINIMUM_CAL_ROUNDS: u64 = 500;
 const MAXIMUM_CAL_ERROR_NS: u64 = 10;
 
 // Don't run the calibration loop for longer than 200ms of wall time.
-const MAXIMUM_CAL_TIME: Duration = Duration::from_millis(200);
+const MAXIMUM_CAL_TIME_NS: u64 = 200 * 1000 * 1000;
 
-#[allow(dead_code)]
 #[derive(Debug)]
 enum ClockType {
     Monotonic(Monotonic),
-    Counter(AtomicCell<u64>, Monotonic, Counter, Calibration),
+    Counter(Monotonic, Counter, Calibration),
     Mock(Arc<Mock>),
 }
 
@@ -180,25 +197,33 @@ pub(crate) struct Calibration {
 }
 
 impl Calibration {
-    pub fn new() -> Calibration {
+    fn new() -> Calibration {
         Calibration {
             ref_time: 0,
             src_time: 0,
             scale_factor: 1,
-            scale_shift: 1,
+            scale_shift: 0,
         }
     }
 
-    fn calibrate(&mut self, reference: &Monotonic, source: &Counter) {
-        // future improvement: read the TSC frequency directly with something like the
-        // code in this PR: https://github.com/hermitcore/uhyve/pull/24
-
-        let mut variance = Variance::default();
-        let deadline = reference.now() + MAXIMUM_CAL_TIME.as_nanos() as u64;
-
+    fn reset_timebases(&mut self, reference: Monotonic, source: &Counter) {
         self.ref_time = reference.now();
         self.src_time = source.now();
+    }
 
+    fn scale_src_to_ref(&self, src_raw: u64) -> u64 {
+        let delta = src_raw.saturating_sub(self.src_time);
+        let scaled = mul_div_po2_u64(delta, self.scale_factor, self.scale_shift);
+        scaled + self.ref_time
+    }
+
+    fn calibrate(&mut self, reference: Monotonic, source: &Counter) {
+        let mut variance = Variance::default();
+        let deadline = reference.now() + MAXIMUM_CAL_TIME_NS;
+
+        self.reset_timebases(reference, source);
+
+        // Each busy loop should spin for 1 microsecond. (1000 nanoseconds)
         let loop_delta = 1000;
         loop {
             // Busy loop to burn some time.
@@ -221,7 +246,7 @@ impl Calibration {
 
             let r_time = reference.now();
             let s_raw = source.now();
-            let s_time = scale_src_to_ref(s_raw, &self);
+            let s_time = self.scale_src_to_ref(s_raw);
             variance.add(s_time as f64 - r_time as f64);
 
             // If we've collected enough samples, check what the mean and mean error are.  If we're
@@ -233,7 +258,7 @@ impl Calibration {
                 let samples = variance.samples();
 
                 if samples > MINIMUM_CAL_ROUNDS
-                    && mwe < MAXIMUM_CAL_ERROR_NS
+                    && mwe < MAXIMUM_CAL_ERROR_NS as f64
                     && mean_error / mean <= 1.0
                 {
                     break;
@@ -242,7 +267,7 @@ impl Calibration {
         }
     }
 
-    fn adjust_cal_ratio(&mut self, reference: &Monotonic, source: &Counter) {
+    fn adjust_cal_ratio(&mut self, reference: Monotonic, source: &Counter) {
         // Overall algorithm: measure the delta between our ref/src_time values and "now" versions
         // of them, calculate the ratio between the deltas, and then find a numerator and
         // denominator to express that ratio such that the denominator is always a power of two.
@@ -255,18 +280,14 @@ impl Calibration {
         // Then, conversion from a raw value simply becomes a multiply and a bit shift instead of a
         // multiply and full-blown divide.
         let ref_end = reference.now();
-        let src_end = source.end();
+        let src_end = source.now();
 
         let ref_d = ref_end.wrapping_sub(self.ref_time);
         let src_d = src_end.wrapping_sub(self.src_time);
 
-        // TODO: we should almost never get a zero here because that would mean denom was greater
-        // than 2^63 which is already a red flag.. but i'm not 100% sure if we can prove it well
-        // enough to simply keep the panic around? gotta think on this
-        let src_d_po2 = src_d.next_power_of_two();
-        if src_d_po2 == 0 {
-            panic!("po2_denom was zero!");
-        }
+        let src_d_po2 = src_d
+            .checked_next_power_of_two()
+            .unwrap_or_else(|| 2_u64.pow(63));
 
         // TODO: lossy conversion back and forth just to get an approximate value, can we do better
         // with integer math? not sure
@@ -293,16 +314,15 @@ impl Clock {
     ///
     /// Support for TSC, etc, are checked at the time of creation, not compile-time.
     pub fn new() -> Clock {
-        let reference = Monotonic::new();
-        let constant_tsc = has_constant_or_better_tsc();
-        let inner = if constant_tsc {
-            let source = Counter::new();
+        let reference = Monotonic::default();
+        let inner = if detection::has_counter_support() {
+            let source = Counter;
             let calibration = GLOBAL_CALIBRATION.get_or_init(|| {
                 let mut calibration = Calibration::new();
-                calibration.calibrate(&reference, &source);
+                calibration.calibrate(reference, &source);
                 calibration
             });
-            ClockType::Counter(AtomicCell::new(0), reference, source, *calibration)
+            ClockType::Counter(reference, source, *calibration)
         } else {
             ClockType::Monotonic(reference)
         };
@@ -332,17 +352,7 @@ impl Clock {
     pub fn now(&self) -> Instant {
         match &self.inner {
             ClockType::Monotonic(monotonic) => Instant(monotonic.now()),
-            ClockType::Counter(last, _, counter, _) => {
-                let now = counter.now();
-                // Update the last timestamp with `now`, if `now` is greater
-                // than the current value.
-                // TODO: replace with `AtomicCell::fetch_max` once `crossbeam-utils` implements it.
-                let last = last.fetch_update(|current| Some(current.max(now))).unwrap();
-                // `fetch_max` always returns the previous value, so we need to
-                // see which is *actually* the max.
-                let actual = std::cmp::max(now, last);
-                self.scaled(actual)
-            }
+            ClockType::Counter(_, counter, _) => self.scaled(counter.now()),
             ClockType::Mock(mock) => Instant(mock.value()),
         }
     }
@@ -353,54 +363,12 @@ impl Clock {
     /// or to be monotonic.  Value can be scaled to reference time by calling either [`scaled`]
     /// or [`delta`].
     ///
-    /// If you need maximum accuracy in your measurements, consider using [`start`] and [`end`].
-    ///
     /// [`scaled`]: Clock::scaled
     /// [`delta`]: Clock::delta
-    /// [`start`]: Clock::start
-    /// [`end`]: Clock::end
     pub fn raw(&self) -> u64 {
         match &self.inner {
             ClockType::Monotonic(monotonic) => monotonic.now(),
-            ClockType::Counter(_, _, counter, _) => counter.now(),
-            ClockType::Mock(mock) => mock.value(),
-        }
-    }
-
-    /// Gets the underlying time from the source clock, specific to starting an operation.
-    ///
-    /// Value is not guaranteed to be in nanoseconds.
-    ///
-    /// Provides the same functionality as [`raw`], but tries to ensure that no extra CPU
-    /// instructions end up executing after the measurement is taken.  Since normal processors are
-    /// typically out-of-order, other operations that logically come before a call to this method
-    /// could be reordered to come after the measurement, thereby skewing the overall time
-    /// measured.
-    ///
-    /// [`raw`]: Clock::raw
-    pub fn start(&self) -> u64 {
-        match &self.inner {
-            ClockType::Monotonic(monotonic) => monotonic.now(),
-            ClockType::Counter(_, _, counter, _) => counter.start(),
-            ClockType::Mock(mock) => mock.value(),
-        }
-    }
-
-    /// Gets the underlying time from the source clock, specific to ending an operation.
-    ///
-    /// Value is not guaranteed to be in nanoseconds.
-    ///
-    /// Provides the same functionality as [`raw`], but tries to ensure that no extra CPU
-    /// instructions end up executing before the measurement is taken.  Since normal processors are
-    /// typically out-of-order, other operations that logically come after a call to this method
-    /// could be reordered to come before the measurement, thereby skewing the overall time
-    /// measured.
-    ///
-    /// [`raw`]: Clock::raw
-    pub fn end(&self) -> u64 {
-        match &self.inner {
-            ClockType::Monotonic(monotonic) => monotonic.now(),
-            ClockType::Counter(_, _, counter, _) => counter.end(),
+            ClockType::Counter(_, counter, _) => counter.now(),
             ClockType::Mock(mock) => mock.value(),
         }
     }
@@ -414,36 +382,53 @@ impl Clock {
     /// Returns an [`Instant`].
     pub fn scaled(&self, value: u64) -> Instant {
         let scaled = match &self.inner {
-            ClockType::Counter(_, _, _, calibration) => scale_src_to_ref(value, &calibration),
+            ClockType::Counter(_, _, calibration) => calibration.scale_src_to_ref(value),
             _ => value,
         };
 
         Instant(scaled)
     }
 
-    /// Calculates the delta between two measurements, and scales to reference time.
+    /// Calculates the delta, in nanoseconds, between two raw measurements.
     ///
-    /// This method is slightly faster when you know you need the delta between two raw
-    /// measurements, or a start/end measurement, than using [`scaled`] for both conversions.
+    /// This method is very similar to [`delta`] but reduces overhead
+    /// for high-frequency measurements that work with nanosecond
+    /// counts internally, as it avoids the conversion of the delta
+    /// into [`Duration`].
     ///
-    /// [`scaled`]: Clock::scaled
-    pub fn delta(&self, start: u64, end: u64) -> Duration {
+    /// [`delta`]: Clock::delta
+    pub fn delta_as_nanos(&self, start: u64, end: u64) -> u64 {
         // Safety: we want wrapping_sub on the end/start delta calculation so that two measurements
         // split across a rollover boundary still return the right result.  However, we also know
         // the TSC could potentially give us different values between cores/sockets, so we're just
         // doing our due diligence here to make sure we're not about to create some wacky duration.
         if end <= start {
-            return Duration::new(0, 0);
+            return 0;
         }
 
         let delta = end.wrapping_sub(start);
-        let scaled = match &self.inner {
-            ClockType::Counter(_, _, _, calibration) => {
+        match &self.inner {
+            ClockType::Counter(_, _, calibration) => {
                 mul_div_po2_u64(delta, calibration.scale_factor, calibration.scale_shift)
             }
             _ => delta,
-        };
-        Duration::from_nanos(scaled)
+        }
+    }
+
+    /// Calculates the delta between two raw measurements.
+    ///
+    /// This method is slightly faster when you know you need the delta between two raw
+    /// measurements, or a start/end measurement, than using [`scaled`] for both conversions.
+    ///
+    /// In code that simply needs access to the whole number of nanoseconds
+    /// between the two measurements, consider [`Clock::delta_as_nanos`]
+    /// instead, which is slightly faster than having to call both this method
+    /// and [`Duration::as_nanos`].
+    ///
+    /// [`scaled`]: Clock::scaled
+    /// [`delta_as_nanos`]: Clock::delta_as_nanos
+    pub fn delta(&self, start: u64, end: u64) -> Duration {
+        Duration::from_nanos(self.delta_as_nanos(start, end))
     }
 
     /// Gets the most recent current time, scaled to reference time.
@@ -453,8 +438,8 @@ impl Clock {
     /// read directly without the need to scale to reference time.
     ///
     /// The upkeep thread must be started in order to update the time.  You can read the
-    /// documentation for [`Builder`] for more information on starting the upkeep thread, as well
-    /// as the details of the "current time" mechanism.
+    /// documentation for [`Upkeep`] for more information on starting the upkeep thread, as
+    /// well as the details of the "current time" mechanism.
     ///
     /// If the upkeep thread has not been started, the return value will be `0`.
     ///
@@ -466,12 +451,15 @@ impl Clock {
         }
     }
 
-    /// Updates the recent current time.
-    ///
-    /// Most callers should use the existing [`Builder`] machinery for spawning a background thread
-    /// to handle upkeep, rather than calling [`upkeep`] directly.
-    pub fn upkeep(value: Instant) {
-        GLOBAL_RECENT.store(value.0);
+    #[cfg(test)]
+    fn reset_timebase(&mut self) -> bool {
+        match &mut self.inner {
+            ClockType::Counter(reference, source, calibration) => {
+                calibration.reset_timebases(*reference, source);
+                true
+            }
+            _ => false,
+        }
     }
 }
 
@@ -486,13 +474,10 @@ impl Clone for ClockType {
     fn clone(&self) -> Self {
         match self {
             ClockType::Mock(mock) => ClockType::Mock(mock.clone()),
-            ClockType::Monotonic(monotonic) => ClockType::Monotonic(monotonic.clone()),
-            ClockType::Counter(last, monotonic, counter, calibration) => ClockType::Counter(
-                AtomicCell::new(last.load()),
-                monotonic.clone(),
-                counter.clone(),
-                *calibration,
-            ),
+            ClockType::Monotonic(monotonic) => ClockType::Monotonic(*monotonic),
+            ClockType::Counter(monotonic, counter, calibration) => {
+                ClockType::Counter(*monotonic, counter.clone(), *calibration)
+            }
         }
     }
 }
@@ -504,7 +489,7 @@ pub fn with_clock<T>(clock: &Clock, f: impl FnOnce() -> T) -> T {
     CLOCK_OVERRIDE.with(|current| {
         let old = current.replace(Some(clock.clone()));
         let result = f();
-        let _ = current.replace(old);
+        current.replace(old);
         result
     })
 }
@@ -516,12 +501,12 @@ pub fn with_clock<T>(clock: &Clock, f: impl FnOnce() -> T) -> T {
 /// recent time is updated.  For example, programs using an asynchronous runtime may prefer to
 /// schedule a task that does the updating, avoiding an extra thread.
 pub fn set_recent(instant: Instant) {
-    GLOBAL_RECENT.store(instant.as_u64());
+    GLOBAL_RECENT.store(instant.0);
 }
 
 #[inline]
 pub(crate) fn get_now() -> Instant {
-    if let Some(instant) = CLOCK_OVERRIDE.with(|clock| clock.borrow().as_ref().map(|c| c.now())) {
+    if let Some(instant) = CLOCK_OVERRIDE.with(|clock| clock.borrow().as_ref().map(Clock::now)) {
         instant
     } else {
         GLOBAL_CLOCK.get_or_init(Clock::new).now()
@@ -537,109 +522,32 @@ pub(crate) fn get_recent() -> Instant {
     // Given that global recent time shouldn't ever be getting _actually_ updated in tests, this
     // should be a reasonable trade-off.
     let recent = GLOBAL_RECENT.load();
-    if recent != 0 {
-        Instant(recent)
-    } else {
+    if recent == 0 {
         get_now()
+    } else {
+        Instant(recent)
     }
 }
 
-#[inline]
-fn scale_src_to_ref(src_raw: u64, cal: &Calibration) -> u64 {
-    let delta = src_raw.saturating_sub(cal.src_time);
-    let scaled = mul_div_po2_u64(delta, cal.scale_factor, cal.scale_shift);
-    scaled + cal.ref_time
-}
-
 #[inline]
 fn mul_div_po2_u64(value: u64, numer: u64, denom: u32) -> u64 {
     // Modified muldiv routine where the denominator has to be a power of two. `denom` is expected
     // to be the number of bits to shift, not the actual decimal value.
-    let mut v: u128 = value as u128;
-    v *= numer as u128;
+    let mut v = u128::from(value);
+    v *= u128::from(numer);
     v >>= denom;
     v as u64
 }
 
-#[allow(dead_code)]
-#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
-fn has_constant_or_better_tsc() -> bool {
-    // All of this special handling for CPU manufacturers, specific family/model combinations, etc,
-    // is derived from the Linux kernel, master branch, as of 2020-05-19.
-
-    // We limit ourselves to Intel, AMD, and Centaur.  There are technically other CPUs that
-    // seemingly support constant or better TSC, but there's no chance in hell I have access to
-    // test that out.  People can reach out if they want support.
-    let cpu_mfg = read_cpuid_mfg();
-    match cpu_mfg.as_str() {
-        "GenuineIntel" => {
-            // All Intel processors from the Core microarchitecture and on.
-            if read_cpuid_family_model() >= 0x60E {
-                return true;
-            }
-        }
-        "AuthenticAMD" => {}
-        "CentaurHauls" => {
-            // VIA Nano and above should at least have constant TSC.
-            if read_cpuid_family_model() >= 0x60F {
-                return true;
-            }
-        }
-        // Unknown/unsupported processor.
-        _ => return false,
-    }
-
-    // Check to make sure we have nonstop TSC + RDTSCP support.
-    read_cpuid_nonstop_tsc() && read_cpuid_rdtscp_support()
-}
-
-#[allow(dead_code)]
-#[cfg(not(all(target_arch = "x86_64", target_feature = "sse2")))]
-fn has_constant_or_better_tsc() -> bool {
-    false
-}
-
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-fn read_cpuid_mfg() -> String {
-    let cpuid = CpuId::new();
-    cpuid
-        .get_vendor_info()
-        .map_or_else(String::new, |vi| vi.as_str().to_owned())
-}
-
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-fn read_cpuid_nonstop_tsc() -> bool {
-    let cpuid = CpuId::new();
-    cpuid
-        .get_advanced_power_mgmt_info()
-        .map_or(false, |efi| efi.has_invariant_tsc())
-}
-
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-fn read_cpuid_rdtscp_support() -> bool {
-    let cpuid = CpuId::new();
-    cpuid
-        .get_extended_processor_and_feature_identifiers()
-        .map_or(false, |efi| efi.has_rdtscp())
-}
-
-#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
-fn read_cpuid_family_model() -> u32 {
-    let cpuid = CpuId::new();
-    cpuid.get_feature_info().map_or(0, |fi| {
-        (fi.family_id() as u32) << 8 | (fi.extended_model_id() as u32) << 4 | fi.model_id() as u32
-    })
-}
-
 #[cfg(test)]
 pub mod tests {
-    use super::{Clock, Monotonic};
+    use super::{Clock, Counter, Monotonic};
     use average::{Merge, Variance};
+    use std::time::{Duration, Instant};
 
     #[cfg(all(target_arch = "wasm32", target_os = "unknown"))]
     mod configure_wasm_tests {
-        // until https://github.com/rustwasm/wasm-bindgen/issues/2571 is resolved
-        // these tests will only run in browsers
+        // Until https://github.com/rustwasm/wasm-bindgen/issues/2571 is resolved, these tests will only run in browsers.
         wasm_bindgen_test::wasm_bindgen_test_configure!(run_in_browser);
     }
 
@@ -650,9 +558,9 @@ pub mod tests {
     )]
     fn test_mock() {
         let (clock, mock) = Clock::mock();
-        assert_eq!(clock.now().as_u64(), 0);
+        assert_eq!(clock.now().0, 0);
         mock.increment(42);
-        assert_eq!(clock.now().as_u64(), 42);
+        assert_eq!(clock.now().0, 42);
     }
 
     #[test]
@@ -662,7 +570,7 @@ pub mod tests {
     )]
     fn test_now() {
         let clock = Clock::new();
-        assert!(clock.now().as_u64() > 0);
+        assert!(clock.now().0 > 0);
     }
 
     #[test]
@@ -675,26 +583,6 @@ pub mod tests {
         assert!(clock.raw() > 0);
     }
 
-    #[test]
-    #[cfg_attr(
-        all(target_arch = "wasm32", target_os = "unknown"),
-        wasm_bindgen_test::wasm_bindgen_test
-    )]
-    fn test_start() {
-        let clock = Clock::new();
-        assert!(clock.start() > 0);
-    }
-
-    #[test]
-    #[cfg_attr(
-        all(target_arch = "wasm32", target_os = "unknown"),
-        wasm_bindgen_test::wasm_bindgen_test
-    )]
-    fn test_end() {
-        let clock = Clock::new();
-        assert!(clock.end() > 0);
-    }
-
     #[test]
     #[cfg_attr(
         all(target_arch = "wasm32", target_os = "unknown"),
@@ -704,7 +592,7 @@ pub mod tests {
         let clock = Clock::new();
         let raw = clock.raw();
         let scaled = clock.scaled(raw);
-        assert!(scaled.as_u64() > 0);
+        assert!(scaled.0 > 0);
     }
 
     #[test]
@@ -713,34 +601,81 @@ pub mod tests {
         all(target_arch = "wasm32", target_os = "unknown"),
         wasm_bindgen_test::wasm_bindgen_test
     )]
+
     fn test_reference_source_calibration() {
-        let clock = Clock::new();
-        let reference = Monotonic::new();
+        let mut clock = Clock::new();
+        let reference = Monotonic::default();
 
         let loops = 10000;
-        let samples = 1024;
 
         let mut overall = Variance::new();
-        let mut deltas = Vec::with_capacity(samples);
-        deltas.reserve(samples);
+        let mut src_samples = [0u64; 1024];
+        let mut ref_samples = [0u64; 1024];
 
         for _ in 0..loops {
-            deltas.clear();
-            for _ in 0..samples {
-                let qstart = clock.now();
-                let rstart = reference.now();
+            // We have to reset the "timebase" of the clock/calibration when testing in this way.
+            //
+            // Since `quanta` is designed around mimicing `Instant`, we care about measuring the _passage_ of time, but
+            // not matching our calculation of wall-clock time to the system's calculation of wall-clock time, in terms
+            // of their absolute values.
+            //
+            // As the system adjusts its clocks over time, whether due to NTP skew, or delays in updating the derived monotonic
+            // time, and so on, our original measurement base from the reference source -- which we use to anchor how we
+            // convert our scaled source measurement into the same reference timebase -- can skew further away from the
+            // current reference time in terms of the rate at which it ticks forward.
+            //
+            // Essentially, what we're saying here is that we want to test the scaling ratio that we generated in
+            // calibration, but not necessarily that the resulting value -- which is meant to be in the same timebase as
+            // the reference -- is locked to the reference itself. For example, if the reference is in nanoseconds, we
+            // want our source to be scaled to nanoseconds, too. We don't care if the system shoves the reference back
+            // and forth via NTP skew, etc... we just need to do enough source-to-reference calibration loops to figure
+            // out what the right amount is to scale the TSC -- since we require an invariant/nonstop TSC -- to get it
+            // to nanoseconds.
+            //
+            // At the risk of saying _too much_, while the delta between `Clock::now` and `Monotonic::now` may grow over
+            // time if the timebases are not reset, we can readily observe in this test that the delta between the
+            // first/last measurement loop for both source/reference are independently close i.e. the ratio by which we
+            // scale the source measurements gets it close, and stays close, to the reference measurements in terms of
+            // the _passage_ of time.
+            clock.reset_timebase();
 
-                deltas.push(rstart.saturating_sub(qstart.as_u64()));
+            for i in 0..1024 {
+                src_samples[i] = clock.now().0;
+                ref_samples[i] = reference.now();
             }
 
-            let local = deltas.iter().map(|i| *i as f64).collect::<Variance>();
+            let is_src_monotonic = src_samples
+                .iter()
+                .map(Some)
+                .reduce(|last, current| last.and_then(|lv| current.filter(|cv| *cv >= lv)))
+                .flatten()
+                .copied();
+            assert_eq!(is_src_monotonic, Some(src_samples[1023]));
+
+            let is_ref_monotonic = ref_samples
+                .iter()
+                .map(Some)
+                .reduce(|last, current| last.and_then(|lv| current.filter(|cv| *cv >= lv)))
+                .flatten()
+                .copied();
+            assert_eq!(is_ref_monotonic, Some(ref_samples[1023]));
+
+            let local = src_samples
+                .iter()
+                .zip(ref_samples.iter())
+                .map(|(s, r)| *s as f64 - *r as f64)
+                .map(|f| f.abs())
+                .collect::<Variance>();
+
             overall.merge(&local);
         }
 
         println!(
-            "reference/source delta: mean={} error={}",
+            "reference/source delta: mean={} error={} mean-var={} samples={}",
             overall.mean(),
-            overall.error()
+            overall.error(),
+            overall.variance_of_mean(),
+            overall.len(),
         );
 
         // If things are out of sync more than 1000ns, something is likely scaled wrong.
@@ -748,41 +683,50 @@ pub mod tests {
     }
 
     #[test]
+    #[cfg_attr(not(feature = "flaky_tests"), ignore)]
     #[cfg_attr(
         all(target_arch = "wasm32", target_os = "unknown"),
         wasm_bindgen_test::wasm_bindgen_test
     )]
-    fn test_reference_self_calibration() {
-        let reference = Monotonic::new();
+    fn measure_source_reference_self_timing() {
+        let source = Counter::default();
+        let reference = Monotonic::default();
 
         let loops = 10000;
-        let samples = 1024;
 
-        let mut overall = Variance::new();
-        let mut deltas = Vec::with_capacity(samples);
-        deltas.reserve(samples);
+        let mut src_deltas = Vec::new();
+        let mut src_samples = [0u64; 100];
 
         for _ in 0..loops {
-            deltas.clear();
-            for _ in 0..samples {
-                let rstart = reference.now();
-                let rend = reference.now();
-
-                deltas.push(rend - rstart);
+            let start = Instant::now();
+            for i in 0..100 {
+                src_samples[i] = source.now();
             }
 
-            let local = deltas.iter().map(|i| *i as f64).collect::<Variance>();
-            overall.merge(&local);
+            src_deltas.push(start.elapsed().as_secs_f64());
         }
 
-        println!(
-            "reference/reference inter-call delta: mean={} error={}",
-            overall.mean(),
-            overall.error()
-        );
+        let mut ref_deltas = Vec::new();
+        let mut ref_samples = [0u64; 100];
 
-        // If things are out of sync more than 1000ns, then I dunno, because our reference is
-        // supposed to be reliable. 😬
-        assert!(overall.mean() < 1000.0);
+        for _ in 0..loops {
+            let start = Instant::now();
+            for i in 0..100 {
+                ref_samples[i] = reference.now();
+            }
+
+            ref_deltas.push(start.elapsed().as_secs_f64());
+        }
+
+        let src_variance = src_deltas.into_iter().collect::<Variance>();
+        let ref_variance = ref_deltas.into_iter().collect::<Variance>();
+
+        let src_variance_ns = Duration::from_secs_f64(src_variance.mean() / 100.0);
+        let ref_variance_ns = Duration::from_secs_f64(ref_variance.mean() / 100.0);
+
+        println!(
+            "source call average: {:?}, reference call average: {:?}",
+            src_variance_ns, ref_variance_ns
+        );
     }
 }
